"""
literal type detection and evaluation
=====================================

Numbers and other none-text-values, e.g. hacked into your application by a user or stored in
a :ref:`configuration file <config-files>` are represented by strings or literals.
An instance of the :class:`Literal` class converts a string into the representing value and type.

A :ref:`evaluable literal <evaluable-literal-formats>` can e.g. be passed
on instantiation through the first (the :paramref:`~Literal.literal_or_value`) argument
of the :class:`Literal` class. The :paramref:`second argument <Literal.value_type>`
specifies the expected type::

    int_literal = Literal("3", int)

Alternatively you could also set the :ref:`evaluable literal string
<evaluable-literal-formats>` after the instantiation directly via the
:attr:`~Literal.value` property setter::

    int_literal = Literal(value_type=int)
    int_literal.value = "3"

The representing numeric value provides the :attr:`~Literal.value` property getter::

    assert int_literal.value == 3
    assert type(int_literal.value) is int

Whenever possible a type conversion will be done, e.g. you can force/restrict the value
to a `float` even if the literal specifies an integer::

    float_literal = Literal("3", float)
    assert float_literal.value = 3.0
    assert type(float_literal.value) is float

So by specifying the type of the literal value within the :paramref:`~Literal.value_type`
argument you restrict a :class:`Literal` instance to a certain/fixed type. Any type/class
can be used::

    list_literal = Literal(value_type=list)
    dict_literal = Literal(value_type=dict)
    datetime_literal = Literal(value_type=datetime.datetime)
    my_class_literal = Literal(value_type=MyClass)

No type info is needed on instantiation when the literal is a evaluable python expression.
For example the following literal gets automatically converted into a datetime object::

    datetime_literal = Literal('(datetime.datetime.now())')

Also here you could alternatively use the :attr:`~Literal.value` property setter::

    date_literal = Literal()
    date_literal.value = '(datetime.date.today())'

.. note::
  The literal string of the last two examples has to start and end with round brackets
  for to mark it as a :ref:`evaluable literal <evaluable-literal-formats>`.

If you instead want to specify a date format literal string then you have
to specify the value type like so::

    date_literal = Literal('2033-12-31', value_type=datetime.date)

As soon as you request the date value from the last `date_literal` examples via
the :attr:`~Literal.value` property getter, the representing/underlying value will be
determined/evaluated and returned::

   literal_value = date_literal.value
   assert literal_value == datetime.date(2033, 12, 31)

The :attr:`~Literal.value` property getter of a :class:`Literal` instance with an applied
type restricting will try to convert the literal to a value of the specified type.
If the evaluation result has not the correct type, then the getter tries the
value conversion with the constructor of the type class. If this fails too then
a ValueError exception will be raised::

    date_literal = Literal(value_type=datetime.date)
    date_literal.value = "invalid-date-literal"
    date_value = date_literal.value             # value getter raises ValueError

All the other supported literal formats are documented at the :attr:`~Literal.value` property.
"""
import datetime
from typing import Any, Callable, Optional, Tuple, Type

from ae.core import DEF_ENCODE_ERRORS, parse_date, try_call, try_eval, try_exec     # type: ignore  # for mypy


__version__ = '0.0.15'


BEG_CHARS = "([{'\""
END_CHARS = ")]}'\""


class Literal:
    """ stores and represents any value, optionally converted from a literal. """

    def __init__(self, literal_or_value: Optional[Any] = None, value_type: Optional[Type] = None, name: str = 'LiT'):
        """ create new Literal instance.

        :param literal_or_value:    initial literal (evaluable string expression) or value of this instance.
        :param value_type:          type of the value of this instance (def=determined latest by/in the
                                    :attr:`~Literal.value` property getter).
        :param name:                name of the literal (only used for debugging/error-message).
        """
        self._name = name
        self._literal_or_value = None
        self._type = None if isinstance(value_type, type(None)) else value_type
        if literal_or_value is not None:
            self.value = literal_or_value

    @property
    def value(self) -> Any:
        """ property representing the value of this Literal instance.

        :setter:    assign literal or a new value; can be either a value literal string or directly
                    the represented/resulting value. If the assigned value is not a string
                    and the value type of this instance got still unspecified then this instance
                    will be restricted to the type of the assigned value.
                    Assigning a None value will be ignored - neither
                    the literal nor the value will change with that!
        :getter:    return the literal value; on the first call the literal will be evaluated
                    (lazy/late) and the value type will be set if still unspecified. Further
                    getter calls will directly return the already converted literal value.

        .. _evaluable-literal-formats:

        If the literal of this :class:`Literal` instance coincide with one of the following
        evaluable formats then the value and the type of the value gets automatically recognized.
        An evaluable formatted literal strings has to start and end with one of the character pairs
        shown in the following table:

        +-------------+------------+------------------------------+
        | starts with | ends with  | evaluation value type        |
        +=============+============+==============================+
        |     (       |     )      | tuple literal or expression  |
        +-------------+------------+------------------------------+
        |     [       |     ]      | list literal                 |
        +-------------+------------+------------------------------+
        |     {       |     }      | dict literal                 |
        +-------------+------------+------------------------------+
        |     '       |     '      | string literal               |
        +-------------+------------+------------------------------+
        |     \"       |     \"      | string literal               |
        +-------------+------------+------------------------------+
        |    '''      |    '''     | code block with return       |
        +-------------+------------+------------------------------+
        |    \"\"\"      |    \"\"\"     | code block with return       |
        +-------------+------------+------------------------------+

        **Other Supported Literals And Values**

        Literals with type restriction to a boolean type are evaluated as python expression.
        This way literal strings like 'True', 'False', '0' and '1' will be correctly recognized
        and converted into a boolean value.

        Literal strings that representing a date value (with type restriction to either
        :class:`datetime.datetime` or :class:`datetime.date`) will be converted with the
        :func:`~ae.core.parse_date` function and should be formatted in one of the
        standard date formats (defined by the :mod:`~ae.core` constants
        :data:`~ae.core.DATE_TIME_ISO` and :data:`~ae.core.DATE_ISO`).

        Literals and values that are not in one of the above formats will finally be passed to
        the constructor of the restricted type class for to try to convert them into their
        representing value.
       """
        check_val = self._literal_or_value
        msg = f"Literal {self._name} with value {check_val!r} "
        if self.type_mismatching_with(check_val):     # first or new late real value conversion/initialization
            try:
                check_val = self._determine_value(check_val)
            except Exception as ex:
                raise ValueError(msg + f"throw exception: {ex}")

        self._chk_val_reset_else_set_type(check_val)
        if check_val is not None:
            if self._type and self.type_mismatching_with(check_val):
                raise ValueError(msg + f"type mismatch: {self._type} != {type(check_val)}")
            self._literal_or_value = check_val

        return self._literal_or_value

    @value.setter
    def value(self, lit_or_val: Any):
        if lit_or_val is not None:
            if isinstance(lit_or_val, bytes) and self._type != bytes:       # if not restricted to bytes
                lit_or_val = lit_or_val.decode('utf-8', DEF_ENCODE_ERRORS)  # ..then convert bytes to string
            self._literal_or_value = lit_or_val     # late evaluation: real value will be checked/converted by getter
            if not self._type and not isinstance(lit_or_val, str):          # set type if unset and no eval
                self._type = type(lit_or_val)

    def append_value(self, item_value: Any) -> Any:
        """ add new item to the list value of this Literal instance (lazy/late self.value getter call function pointer).

        :param item_value:  value of the item to be appended to the value of this Literal instance.
        :return:            the value (==list) of this Literal instance.

        This method gets e.g. used by the :class:`~.console.ConsoleApp` method
        :meth:`~.console.ConsoleApp.add_option` for to have a function pointer to this
        literal value with lazy/late execution of the value getter (value.append cannot be used in this case
        because the list could have be changed before it get finally read/used).

        .. note::
           This method calls the append method of the value object and will therefore
           only work if the value is of type :class:`list` (or a compatible type).
        """
        self.value.append(item_value)
        return self.value

    def convert_value(self, lit_or_val: Any) -> Any:
        """ set/change the literal/value of this :class:`Literal` instance and return the represented value.

        :param lit_or_val:  the new value to be set.
        :return:            the final/converted value of this Literal instance.

        This method gets e.g. used by the :class:`~.console.ConsoleApp` method
        :meth:`~.console.ConsoleApp.add_option` for to have a function pointer
        for to let the ArgumentParser convert a configuration option literal into the
        represented value.
        """
        self.value = lit_or_val
        return self.value

    def type_mismatching_with(self, value: Any) -> bool:
        """ check if this literal instance would reject the passed value because of type mismatch.

        :param value:       new literal value.
        :return:            True if the passed value would have an type mismatch or if literal type is still not set,
                            else False.
        """
        return self._type != type(value)

    def _determine_value(self, lit_or_val: Any) -> Any:
        """ check passed value if it is still a literal determine the represented value.

        :param lit_or_val:  new literal value or the representing literal string.
        :return:            determined/converted value or self._lit_or_val if value could not be recognized/converted.
        """
        if isinstance(lit_or_val, str):
            func, eval_expr = self._evaluable_literal(lit_or_val)
            if func:
                lit_or_val = self._chk_val_reset_else_set_type(func(eval_expr))

        if self._type:
            if self.type_mismatching_with(lit_or_val) and isinstance(lit_or_val, str):
                if self._type == bool:
                    lit_or_val = bool(try_eval(lit_or_val))
                elif self._type in (datetime.date, datetime.datetime):
                    lit_or_val = parse_date(lit_or_val, ret_date=self._type == datetime.date)
                lit_or_val = self._chk_val_reset_else_set_type(lit_or_val)

            if self.type_mismatching_with(lit_or_val):          # finally try type conversion with type constructor
                lit_or_val = self._chk_val_reset_else_set_type(
                    try_call(self._type, lit_or_val, ignored_exceptions=(TypeError,)))  # ignore int(None) exception

        return lit_or_val

    @staticmethod
    def _evaluable_literal(literal: str) -> Tuple[Optional[Callable], Optional[str]]:
        """ check evaluable format of literal and possibly return appropriate evaluation function and stripped literal.

        :param literal:     string to be checked if it is in the
                            :ref:`evaluable literal format <evaluable-literal-formats>` and if
                            it has to be stripped.
        :return:            tuple of evaluation/execution function and the (optionally stripped) literal
                            string (removed triple high-commas on expression/code-blocks) - if
                            :paramref:`~._evaluable_literal.literal` is in one of the supported
                            :ref:`evaluable literal formats <evaluable-literal-formats>` - else the tuple
                            (None, <empty string>).
        """
        func = None
        ret = ''
        if (literal.startswith("'''") and literal.endswith("'''")) \
                or (literal.startswith('"""') and literal.endswith('"""')):
            func = try_exec
            ret = literal[3:-3]                                             # code block
        elif literal and literal[0] in BEG_CHARS and BEG_CHARS.find(literal[0]) == END_CHARS.find(literal[-1]):
            # below expression used before for elif above but pylint doesn't like more than 5 boolean expression (R0916)
            # elif (literal.startswith("[") and literal.endswith("]")) \ ...
            #    or (literal.startswith("{") and literal.endswith("}")) \
            #    or (literal.startswith("(") and literal.endswith(")")) \
            #    or (literal.startswith("'") and literal.endswith("'")) \
            #    or (literal.startswith('"') and literal.endswith('"')):
            func = try_eval
            ret = literal                                                   # expression/list/dict/tuple/str/... literal
        return func, ret

    def _chk_val_reset_else_set_type(self, value: Any) -> Any:
        """ reset and return passed value if is None, else determine value type and set type (if not already set).

        :param value:       just converted new literal value for to be checked and if ok used to set an unset type.
        :return:            passed value or the stored literal/value if passed value is None.
        """
        if value is None:
            value = self._literal_or_value  # literal evaluation failed, therefore reset to try with type conversion
        elif not self._type and value is not None:
            self._type = type(value)
        return value



""" common setup for root and portions (modules or sub-packages) of the ae namespace package.

# THIS FILE IS EXCLUSIVELY MAINTAINED IN THE AE ROOT PACKAGE. ANY CHANGES SHOULD BE DONE THERE.
# All changes will be deployed automatically to all the portions of this namespace package.

This file get run by each portion of this namespace package for builds (sdist/bdist_wheels)
and installation (install); also gets imported by the root package (for the globals defined
here) for documentation build (docs/conf.py), common file deploys and commit preparation.
"""
import glob
import os
import re
import setuptools
from typing import Dict, List


PT_PKG: str = 'sub-package'         #: sub-package portion type
PT_MOD: str = 'module'              #: module portion type

version_patch_parser = re.compile(r"(^__version__ = ['\"]\d*[.]\d*[.])(\d+)([a-z]*['\"])", re.MULTILINE)


def bump_code_file_patch_number(file_name: str) -> str:
    """ read code file version and then increment the patch number by one and write the code file back. """
    if not os.path.exists(file_name):
        return f"Not existing file {file_name}"
    content = file_content(file_name)
    if not content:
        return f"Empty file {file_name}"
    content, replaced = version_patch_parser.subn(lambda m: m.group(1) + str(int(m.group(2)) + 1) + m.group(3), content)
    if replaced != 1:
        return f"Variable __version__ found {replaced} times in {file_name}"
    with open(file_name, 'w') as fp:
        fp.write(content)
    return ""


def code_file_version(file_name: str) -> str:
    """ read version of Python code file - from __version__ module variable initialization. """
    content = file_content(file_name)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M)
    if not version_match:
        raise FileNotFoundError(f"Unable to find version string within {file_name}")
    return version_match.group(1)


def determine_package_vars(portion_root_path: str, portion_type: str = PT_MOD, portion_end: str = '.py'
                           ) -> Dict[str, str]:
    """ determine vars of a ae namespace package portion (and if it is either a module or a sub-package). """
    if os.path.exists(portion_root_path):                   # run/imported by portion repository
        search_module = portion_type == PT_MOD
        files = [fn for fn in glob.glob(os.path.join(portion_root_path, '*' + portion_end)) if '__' not in fn]
        if len(files) > 1:
            raise RuntimeError(f"More than one {portion_type} found: {files}")
        if len(files) == 0:
            if not search_module:
                raise RuntimeError(f"Neither module nor sub-package found in package path {portion_root_path}")
            return determine_package_vars(portion_root_path, PT_PKG, os.path.sep)
        portion_name = os.path.split(files[0][:-len(portion_end)])[1]
    else:                                                   # imported by namespace root repo
        portion_type = ''
        portion_name = "{portion-name}"

    p_vars = dict()
    p_vars['portion_type'] = portion_type
    p_vars['portion_name'] = portion_name
    p_vars['portion_file_name'] = portion_name + (os.path.sep + '__init__.py' if portion_type == PT_PKG else '.py')
    p_vars['portion_file_path'] = os.path.abspath(os.path.join(portion_root_path, p_vars['portion_file_name']))
    p_vars['package_name'] = f"{namespace_name}_{portion_name}"
    p_vars['pip_name'] = f"{namespace_name}-{portion_name}"
    p_vars['import_name'] = f"{namespace_name}.{portion_name}"
    p_vars['package_version'] = code_file_version(p_vars['portion_file_path']) if portion_type else 'x.y.z'
    p_vars['root_version'] = 'un.kno.wn' if portion_type else code_file_version(os.path.join(setup_path, 'setup.py'))
    p_vars['repo_url_root'] = f"https://gitlab.com/{namespace_name}-group/"

    return p_vars


def determine_setup_path() -> str:
    """ check if setup.py got called from portion root or from docs/RTD root. """
    cwd = os.getcwd()
    if os.path.exists('setup.py'):      # local build
        return cwd
    if os.path.exists('conf.py'):       # RTD build
        return os.path.abspath('..')
    raise RuntimeError(f"Neither setup.py nor conf.py found in current working directory {cwd}")


def file_content(file_name: str) -> str:
    """ returning content of the file specified by file_name arg as string. """
    with open(file_name) as fp:
        return fp.read()


def patch_templates(patch_vars: Dict[str, str], exclude_folder: str = '') -> List[str]:
    """ convert ae namespace package templates found in the cwd or underneath (except excluded) to the final files. """
    patched = list()
    for fn in glob.glob('**/*.*' + template_extension, recursive=True):
        if not exclude_folder or not fn.startswith(exclude_folder + os.path.sep):
            content = file_content(fn).format(**patch_vars)
            with open(fn[:-len(template_extension)], 'w') as fp:
                fp.write(content)
            patched.append(fn)
    return patched


namespace_name = 'ae'
portions_common_root_path = 'portions_common_root'
template_extension = '.tpl'
setup_path = determine_setup_path()
portion_path = os.path.join(setup_path, namespace_name)
package_vars = determine_package_vars(portion_path)
package_name = package_vars['package_name']
repo_url_root = package_vars['repo_url_root']

requirements_file = os.path.join(setup_path, 'requirements.txt')
if os.path.exists(requirements_file):
    dev_require = [_ for _ in file_content(requirements_file).strip().split('\n')
                   if not _.startswith('#')]
else:
    dev_require = ['pytest', 'pytest-cov']
docs_require = [_ for _ in dev_require if _.startswith('sphinx_')]
tests_require = [_ for _ in dev_require if _.startswith('pytest')]
portions_package_names = [_ for _ in dev_require if _.startswith('ae_')]

# provide additional package info for root package templates
package_vars['portions_common_root_path'] = portions_common_root_path
package_vars['portions_pypi_refs_md'] = "\n".join(
    f'* [{_}](https://pypi.org/project/{_} "ae namespace portion {_}")'
    for _ in portions_package_names)                        # used in ./README.md.tpl
namespace_len = len(namespace_name)
package_vars['portions_import_names'] = ("\n" + " " * 4).join(
    _[:namespace_len] + '.' + _[namespace_len + 1:]
    for _ in portions_package_names)                        # used in docs/index.rst.tpl


if __name__ == "__main__":
    setuptools.setup(
        name=package_name,              # pip install name (not the import package name)
        version=package_vars['package_version'],
        author="Andi Ecker",
        author_email="aecker2@gmail.com",
        description=package_name + " portion of python application environment namespace package",
        long_description=file_content("README.md"),
        long_description_content_type="text/markdown",
        url=f"{repo_url_root}{package_name}",
        # don't needed for native/implicit namespace packages: namespace_packages=['ae'],
        # packages=setuptools.find_packages(),
        packages=setuptools.find_namespace_packages(include=[namespace_name]),  # find ae namespace portions
        python_requires=">=3.6",
        extras_require={
            'docs': docs_require,
            'tests': tests_require,
            'dev': docs_require + tests_require,
        },
        classifiers=[
            "Development Status :: 1 - Planning",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Operating System :: OS Independent",
            "Topic :: Software Development :: Libraries :: Application Frameworks",
        ],
        keywords=[
            'productivity',
            'application',
            'environment',
            'configuration',
            'development',
        ]
    )
